from pathlib import Path
from typing import Any, Dict, Iterator, List, MutableMapping, Optional, cast
from typing_extensions import override

import msgpack
from cookit import FileCacheManager
from cookit.loguru.common import logged_suppress
from nonebot_plugin_alconna.uniseg import UniMessage

from .config import config

# dumper and validators just works with simple standard UniMessages
# but it's enough for this plugin, maybe ¯\_(ツ)_/¯


class MessageCacheManager(MutableMapping[str, Optional[List[UniMessage]]]):
    def __init__(
        self,
        cache_dir: Path,
        max_size: Optional[int] = None,
        ttl: Optional[int] = None,
    ) -> None:
        super().__init__()
        self.cache = FileCacheManager(cache_dir, max_size=max_size, ttl=ttl)

    @override
    def __getitem__(self, key: str) -> Optional[List[UniMessage]]:
        data = self.cache[key]
        with logged_suppress("Failed to read message cache"):
            unpacked: List[List[Dict[str, Any]]] = msgpack.unpackb(data)
            return [UniMessage.load(x) for x in unpacked]
        return None

    @override
    def __setitem__(self, key: str, value: Optional[List[UniMessage]]) -> None:
        if not value:
            raise ValueError("value cannot be empty")
        data = None
        with logged_suppress("Failed to dump message cache"):
            dumped = [x.dump(media_save_dir=False) for x in value]
            data = cast(bytes, msgpack.packb(dumped))
        if data:
            self.cache[key] = data

    @override
    def __delitem__(self, key: str) -> None:
        self.cache.__delitem__(key)

    @override
    def __iter__(self) -> Iterator[str]:
        return self.cache.__iter__()

    @override
    def __len__(self) -> int:
        return self.cache.__len__()

    @override
    def __contains__(self, key: Any) -> bool:
        return self.cache.__contains__(key)


CACHE_DIR = Path.cwd() / "data" / "YetAnotherPicSearch" / "cache"
msg_cache = MessageCacheManager(
    CACHE_DIR,
    max_size=config.cache_expire * 100,
    ttl=config.cache_expire * 24 * 60 * 60,
)

# delete old cache
for _it in (x for x in Path.cwd().glob("pic_search_cache*") if x.is_file()):
    _it.unlink()



import math
from typing import Any, Dict

from httpx import AsyncClient
from nonebot_plugin_alconna.uniseg import UniMessage
from PicImageSearch import TraceMoe

from ..config import config
from ..registry import SearchFunctionReturnType
from ..utils import async_lock, combine_message, handle_img


@async_lock()
async def whatanime_search(
    file: bytes,
    client: AsyncClient,
    _: str,
) -> SearchFunctionReturnType:
    whatanime = TraceMoe(client=client)
    res = await whatanime.search(file=file)
    if res and res.raw:
        time = res.raw[0].From
        minutes = math.floor(time / 60)
        seconds = math.floor(time % 60)
        time_str = f"{minutes:02d}:{seconds:02d}"

        if res.raw[0].isAdult:
            thumbnail = await handle_img(
                res.raw[0].cover_image,
                config.hide_img or config.hide_img_when_whatanime_r18,
            )
        else:
            thumbnail = await handle_img(
                res.raw[0].cover_image,
            )

        chinese_title = res.raw[0].title_chinese
        native_title = res.raw[0].title_native

        start_date = date_to_str(res.raw[0].start_date)
        end_date = ""
        if (end_date_year := res.raw[0].end_date["year"]) and end_date_year > 0:
            end_date = date_to_str(res.raw[0].end_date)
        episode = res.raw[0].episode or 1
        res_list = [
            f"WhatAnime ({res.raw[0].similarity}%)",
            f"该截图出自第 {episode} 集的 {time_str}",
            thumbnail,
            native_title,
            chinese_title if chinese_title != native_title else "",
            f"类型：{res.raw[0].type}-{res.raw[0].format}",
            f"开播：{start_date}",
            f"完结：{end_date}" if end_date else "",
        ]
        return [combine_message(res_list)]

    return [UniMessage.text("WhatAnime 暂时无法使用")]


def date_to_str(date: Dict[str, Any]) -> str:
    return f"{date['year']}-{date['month']}-{date['day']}"



from typing import List

from httpx import AsyncClient
from nonebot_plugin_alconna.uniseg import UniMessage
from PicImageSearch import Yandex
from PicImageSearch.model import YandexResponse

from ..registry import SearchFunctionReturnType, search_function
from ..utils import async_lock, combine_message, handle_img, shorten_url


@search_function("yandex")
@async_lock()
async def yandex_search(
    file: bytes,
    client: AsyncClient,
    _: str,
) -> SearchFunctionReturnType:
    yandex = Yandex(client=client)
    if res := await yandex.search(file=file):
        return await search_result_filter(res)
    return [UniMessage.text("Yandex 暂时无法使用")]


async def search_result_filter(res: YandexResponse) -> List[UniMessage]:
    url = await shorten_url(res.url)
    if not res.raw:
        return [UniMessage.text(f"Yandex 搜索结果为空\n搜索页面：{url}")]

    thumbnail = await handle_img(res.raw[0].thumbnail)
    res_list = [
        "Yandex 搜索结果",
        thumbnail,
        res.raw[0].size,
        res.raw[0].title,
        res.raw[0].source,
        res.raw[0].content,
        f"来源：{res.raw[0].url}",
        f"搜索页面：{url}",
    ]
    return [combine_message(res_list)]



# ruff: noqa: E402, N999

from nonebot.plugin import PluginMetadata, inherit_supported_adapters, require

require("nonebot_plugin_waiter")
require("nonebot_plugin_alconna")

from . import __main__ as __main__
from .config import ConfigModel, config
from .data_source import load_search_func

load_search_func()

__version__ = "2.0.1.post1"
__plugin_meta__ = PluginMetadata(
    name="YetAnotherPicSearch",
    description="基于 NoneBot2 及 PicImageSearch 的另一个 NoneBot 搜图插件",
    usage=f"使用指令 `{config.search_keyword} -h` 查看帮助",
    type="application",
    homepage="https://github.com/lgc-NB2Dev/YetAnotherPicSearch",
    config=ConfigModel,
    supported_adapters=inherit_supported_adapters(
        "nonebot_plugin_waiter",
        "nonebot_plugin_alconna",
    ),
    extra={},
)



import asyncio
import operator
import re
from contextlib import suppress
from difflib import SequenceMatcher
from functools import wraps
from io import BytesIO
from pathlib import Path
from typing import (
    Awaitable,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    TypeVar,
    Union,
    overload,
)
from typing_extensions import ParamSpec

import arrow
from cachetools import TTLCache
from cookit.loguru import logged_suppress
from httpx import URL, AsyncClient, HTTPStatusError, InvalidURL
from nonebot.adapters.onebot.v11 import Bot
from nonebot.matcher import current_bot, current_event, current_matcher
from nonebot_plugin_alconna.uniseg import Image as ImageSeg, UniMessage, image_fetch
from PicImageSearch.model.ehentai import EHentaiItem, EHentaiResponse
from PIL import Image
from pyquery import PyQuery
from shelved_cache import cachedasyncmethod
from tenacity import TryAgain, retry, stop_after_attempt, stop_after_delay

from .config import config
from .nhentai_model import NHentaiItem, NHentaiResponse

T = TypeVar("T")
P = ParamSpec("P")


DEFAULT_HEADERS = {
    "User-Agent": (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
        "AppleWebKit/537.36 (KHTML, like Gecko) "
        "Chrome/99.0.4844.82 Safari/537.36"
    ),
}


def post_image_process(file: bytes) -> bytes:
    im = Image.open(BytesIO(file))
    if (im.format == "WEBP") or getattr(im, "is_animated", False):
        with BytesIO() as output:
            im.save(output, "PNG")
            return output.getvalue()
    return file


@retry(stop=(stop_after_attempt(3) | stop_after_delay(30)), reraise=True)
async def get_image_bytes_by_url(url: str, cookies: Optional[str] = None) -> bytes:
    _url = URL(url)
    referer = f"{_url.scheme}://{_url.host}/"
    headers = (
        DEFAULT_HEADERS
        if _url.host.endswith("qpic.cn")
        else {"Referer": referer, **DEFAULT_HEADERS}
    )
    async with AsyncClient(
        headers=headers,
        cookies=parse_cookies(cookies),
        proxies=config.proxy,
        follow_redirects=True,
    ) as session:
        resp = await session.get(url)
        if resp.status_code == 404:
            resp.raise_for_status()
            raise Exception  # NoReturn for sure, just make linter know
        if resp.status_code >= 400 or len(resp.content) == 0:
            raise TryAgain
        return post_image_process(resp.content)


async def get_image_from_seg(seg: ImageSeg) -> bytes:
    with suppress(ValueError):
        return seg.raw_bytes
    if seg.path:
        return Path(seg.path).read_bytes()
    if file := await image_fetch(
        current_event.get(),
        current_bot.get(),
        current_matcher.get().state,
        seg,
    ):
        return file
    raise ValueError("Cannot get image")


async def handle_img(
    url: str,
    hide_img: bool = config.hide_img,
    cookies: Optional[str] = None,
) -> UniMessage:
    if not hide_img:
        with logged_suppress("Failed to get image", HTTPStatusError):
            return UniMessage.image(raw=await get_image_bytes_by_url(url, cookies))
    return UniMessage.text(f"预览图链接：{url}")


@cachedasyncmethod(lambda _: TTLCache(1, 300))
async def get_bot_friend_list(bot: Bot) -> List[int]:
    friend_list = await bot.get_friend_list()
    return [i["user_id"] for i in friend_list]


def handle_source(source: str) -> str:
    return (
        source.replace("www.pixiv.net/en/artworks", "www.pixiv.net/artworks")
        .replace(
            "www.pixiv.net/member_illust.php?mode=medium&illust_id=",
            "www.pixiv.net/artworks/",
        )
        .replace("http://", "https://")
    )


def parse_source(resp_text: str, host: str) -> Optional[str]:
    if host in ["danbooru.donmai.us", "gelbooru.com"]:
        return PyQuery(resp_text)(".image-container").attr("data-normalized-source")  # type: ignore

    if host in ["yande.re", "konachan.com"]:
        source = PyQuery(resp_text)("#post_source").attr("value")
        return source or PyQuery(resp_text)('a[href^="/pool/show/"]').text()  # type: ignore

    return ""


async def get_source(url: str) -> str:
    if not url:
        return ""

    _url = get_valid_url(url)
    if not _url:
        return ""

    host = _url.host
    headers = None if host == "danbooru.donmai.us" else DEFAULT_HEADERS
    async with AsyncClient(
        headers=headers,
        proxies=config.proxy,
        follow_redirects=True,
    ) as session:
        resp = await session.get(url)
        if resp.status_code >= 400:
            return ""

        source = parse_source(resp.text, host)
        if source and get_valid_url(source):
            return handle_source(source)

    return source or ""


def confuse_url(url: str) -> str:
    return next(
        (
            url.replace("//", "// ").replace(host, host.replace(".", ". "))
            for host in config.to_confuse_urls
            if host in url
        ),
        url,
    )


async def shorten_url(url: str) -> str:
    pid_search = re.compile(
        r"(?:pixiv.+(?:illust_id=|artworks/)|/img-original/img/(?:\d+/){6})(\d+)",
    )
    if pid_match := pid_search.search(url):
        return confuse_url(f"https://pixiv.net/i/{pid_match[1]}")

    uid_search = re.compile(r"pixiv.+(?:member\.php\?id=|users/)(\d+)")
    if uid_match := uid_search.search(url):
        return confuse_url(f"https://pixiv.net/u/{uid_match[1]}")

    host = URL(url).host
    if host == "danbooru.donmai.us":
        return confuse_url(url.replace("/post/show/", "/posts/"))

    if host in [
        "e-hentai.org",
        "exhentai.org",
        "graph.baidu.com",
        "nhentai.net",
        "www.google.com",
        "yandex.com",
    ]:
        flag = len(url) > 1024
        async with AsyncClient(headers=DEFAULT_HEADERS) as session:
            if not flag:
                resp = await session.post("https://yww.uy/shorten", json={"url": url})
                if resp.status_code < 400:
                    return resp.json()["url"]  # type: ignore
                flag = True
            if flag:
                resp = await session.post(
                    "https://www.shorturl.at/shortener.php",
                    data={"u": url},
                )
                if resp.status_code < 400:
                    final_url = PyQuery(resp.text)("#shortenurl").attr("value")
                    return f"https://{final_url}"

    return confuse_url(url)


def parse_cookies(cookies_str: Optional[str] = None) -> Dict[str, str]:
    cookies_dict: Dict[str, str] = {}
    if cookies_str:
        for line in cookies_str.split(";"):
            key, value = line.strip().split("=", 1)
            cookies_dict[key] = value
    return cookies_dict


def async_lock(freq: float = 1):
    def decorator(func: Callable[P, Awaitable[T]]):
        lock = asyncio.Lock()
        last_call_time: Optional[arrow.Arrow] = None

        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            nonlocal last_call_time
            async with lock:
                elapsed_time = arrow.now() - (
                    last_call_time or arrow.now().shift(seconds=-freq)
                )
                if elapsed_time.total_seconds() < freq:
                    await asyncio.sleep(freq - elapsed_time.total_seconds())
                result = await func(*args, **kwargs)
                last_call_time = arrow.now()
                return result

        return wrapper

    return decorator


def preprocess_search_query(query: str) -> str:
    query = re.sub(r"●|・|~|～|〜|、|×|:::|\s+-\s+|\[中国翻訳]", " ", query)
    # 去除独立的英文、日文、中文字符，但不去除带连字符的
    for i in [
        r"\b[A-Za-z]\b",
        r"\b[\u4e00-\u9fff]\b",
        r"\b[\u3040-\u309f\u30a0-\u30ff]\b",
    ]:
        query = re.sub(rf"(?<!-){i}(?!-)", "", query)

    return query.strip()


@overload
def filter_results_with_ratio(
    res: EHentaiResponse,
    title: str,
) -> List[EHentaiItem]: ...
@overload
def filter_results_with_ratio(
    res: NHentaiResponse,
    title: str,
) -> List[NHentaiItem]: ...
def filter_results_with_ratio(
    res: Union[EHentaiResponse, NHentaiResponse],
    title: str,
) -> Union[List[EHentaiItem], List[NHentaiItem]]:
    raw_with_ratio = [
        (i, SequenceMatcher(lambda x: x == " ", title, i.title).ratio())
        for i in res.raw
    ]
    raw_with_ratio.sort(key=operator.itemgetter(1), reverse=True)

    if filtered := [i[0] for i in raw_with_ratio if i[1] > 0.65]:
        return filtered  # type: ignore
    return [i[0] for i in raw_with_ratio]  # type: ignore


def get_valid_url(url: str) -> Optional[URL]:
    with suppress(InvalidURL):
        url_obj = URL(url)
        if url_obj.host:
            return url_obj
    return None


def combine_message(
    msg_list: Iterable[Union[UniMessage, str, None]],
    join: Optional[str] = "\n",
) -> UniMessage:
    msg = UniMessage()
    for i, it in enumerate(msg_list):
        if not it:
            continue
        if join and i != 0:
            msg += join
        msg += it
    return msg



from httpx import AsyncClient
from nonebot_plugin_alconna.uniseg import UniMessage
from PicImageSearch import BaiDu

from ..registry import SearchFunctionReturnType, search_function
from ..utils import async_lock, combine_message, handle_img, shorten_url


@search_function("baidu")
@async_lock()
async def baidu_search(
    file: bytes,
    client: AsyncClient,
    _: str,
) -> SearchFunctionReturnType:
    baidu = BaiDu(client=client)
    res = await baidu.search(file=file)
    _url = await shorten_url(res.url)
    if not res.raw:
        return [UniMessage.text(f"Baidu 搜索结果为空\n搜索页面：{_url}")]
    thumbnail = await handle_img(res.raw[0].thumbnail)
    res_list = [
        "Baidu 搜索结果",
        thumbnail,
        res.raw[0].url,
        f"搜索页面：{_url}",
    ]
    return [combine_message(res_list)]



from typing import List

from httpx import AsyncClient
from nonebot_plugin_alconna.uniseg import UniMessage
from PicImageSearch import Iqdb

from ..config import config
from ..registry import SearchFunctionReturnType, search_function
from ..utils import (
    async_lock,
    combine_message,
    get_source,
    get_valid_url,
    handle_img,
    shorten_url,
)
from .ascii2d import ascii2d_search


@search_function("iqdb")
@async_lock()
async def iqdb_search(
    file: bytes,
    client: AsyncClient,
    _: str,
) -> SearchFunctionReturnType:
    iqdb = Iqdb(client=client)
    res = await iqdb.search(file=file)
    if not res.raw:
        return (
            [UniMessage.text("Iqdb 暂时无法使用，自动使用 Ascii2D 进行搜索")],
            ascii2d_search,
        )

    final_res: List[UniMessage] = []
    # 如果遇到搜索结果相似度低的情况，去除第一个只有提示信息的空结果
    low_acc = False
    if res.raw[0].content == "No relevant matches":
        low_acc = True
        res.raw.pop(0)
    selected_res = res.raw[0]
    hide_img = config.hide_img or (low_acc and config.hide_img_when_low_acc)

    # 优先取 danbooru 或 yande.re
    danbooru_res_list = [i for i in res.raw if i.source == "Danbooru"]
    yandere_res_list = [i for i in res.raw if i.source == "yande.re"]
    if danbooru_res_list:
        selected_res = danbooru_res_list[0]
    elif yandere_res_list:
        selected_res = yandere_res_list[0]

    thumbnail = await handle_img(selected_res.thumbnail, hide_img)
    source = await get_source(selected_res.url)
    if source:
        if get_valid_url(source):
            source = await shorten_url(source)
        source = f"来源：{source}"
    res_list = [
        f"Iqdb ({selected_res.similarity}%)",
        thumbnail,
        await shorten_url(selected_res.url),
        source,
        f"搜索页面：{res.url}",
    ]
    final_res.append(combine_message(res_list))

    if low_acc and config.auto_use_ascii2d:
        final_res.append(
            UniMessage.text(
                f"相似度 {selected_res.similarity}% 过低，自动使用 Ascii2D 进行搜索",
            ),
        )
        return final_res, ascii2d_search

    return final_res



import base64
from typing import List

from httpx import AsyncClient
from nonebot_plugin_alconna.uniseg import UniMessage
from PicImageSearch import Google
from PicImageSearch.model import GoogleResponse

from ..registry import SearchFunctionReturnType, search_function
from ..utils import async_lock, combine_message, shorten_url


@search_function("google")
@async_lock()
async def google_search(
    file: bytes,
    client: AsyncClient,
    _: str,
) -> SearchFunctionReturnType:
    google = Google(client=client)
    if res := await google.search(file=file):
        return await search_result_filter(res)
    return [UniMessage.text("Google 暂时无法使用")]


async def search_result_filter(res: GoogleResponse) -> List[UniMessage]:
    url = await shorten_url(res.url)
    if not res.raw:
        return [UniMessage.text(f"Google 搜索结果为空\n搜索页面：{url}")]

    selected_res = next((i for i in res.raw if i.thumbnail), res.raw[0])
    if not selected_res.thumbnail:
        return [UniMessage.text(f"Google 搜索结果为空\n搜索页面：{url}")]

    thumbnail = UniMessage.image(
        raw=base64.b64decode(selected_res.thumbnail.split(",", 1)[1]),
    )
    res_list = [
        "Google 搜索结果",
        thumbnail,
        selected_res.title,
        f"来源：{selected_res.url}",
        f"搜索页面：{url}",
    ]
    return [combine_message(res_list)]
